# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog
~~~~~~~

Simple client for A10 load balancers.
"""

__version__ = '1.1.0'
